/*
 * Copyright 2014-2019 JetBrains s.r.o and contributors. Use of this source code is governed by the Apache 2.0 license.
 */

package io.ktor.server.application

import io.ktor.events.*
import io.ktor.server.engine.*
import io.ktor.util.*
import io.ktor.util.logging.*
import kotlinx.coroutines.*
import kotlin.coroutines.*

public class ApplicationPropertiesBuilder(
    public val environment: ApplicationEnvironment
) {

    internal val modules: MutableList<Application.() -> Unit> = mutableListOf()

    /**
     * Paths to wait for application reload
     */
    public var watchPaths: List<String> = listOf(WORKING_DIRECTORY_PATH)

    /**
     * Application's root path (prefix, context path in servlet container).
     */
    public var rootPath: String = ""

    /**
     * Indicates if development mode is enabled.
     */
    public var developmentMode: Boolean = PlatformUtils.IS_DEVELOPMENT_MODE

    /**
     * Install application module
     */
    public fun module(body: Application.() -> Unit) {
        modules.add(body)
    }

    internal fun build(): ApplicationProperties =
        ApplicationProperties(environment, modules, watchPaths, rootPath, developmentMode)
}

public class ApplicationProperties(
    public val environment: ApplicationEnvironment,
    internal val modules: MutableList<Application.() -> Unit>,
    internal val watchPaths: List<String>,
    public val rootPath: String,
    public val developmentMode: Boolean = PlatformUtils.IS_DEVELOPMENT_MODE
)

public fun applicationProperties(
    environment: ApplicationEnvironment = applicationEnvironment {},
    block: ApplicationPropertiesBuilder.() -> Unit = {}
): ApplicationProperties {
    return ApplicationPropertiesBuilder(environment).apply(block).build()
}

/**
 * Represents configured and running web application, capable of handling requests.
 * It is also the application coroutine scope that is cancelled immediately at application stop so useful
 * for launching background coroutines.
 *
 * @param environment Instance of [ApplicationEnvironment] describing environment this application runs in
 */
@KtorDsl
public class Application(
    environment: ApplicationEnvironment,
    developmentMode: Boolean,
    public var rootPath: String,
    public val monitor: Events
) : ApplicationCallPipeline(developmentMode, environment), CoroutineScope {

    private val applicationJob = SupervisorJob(environment.parentCoroutineContext[Job])

    override val coroutineContext: CoroutineContext = environment.parentCoroutineContext + applicationJob

    /**
     * Called by [ApplicationEngine] when [Application] is terminated
     */
    @Suppress("DEPRECATION")
    public fun dispose() {
        applicationJob.cancel()
        uninstallAllPlugins()
    }
}

/**
 * Convenience property to access log from application
 */
public val Application.log: Logger get() = environment.log
